-----------------------------------------------------------------------------
-- Default:	Defaulting
-- 
-- Part of `Typing Haskell in Haskell', version of September 1, 1999
-- Copyright (c) Mark P Jones and the Oregon Graduate Institute
-- of Science and Technology, 1999
-- 
-- This program is distributed as Free Software under the terms
-- in the file "License" that is included in the distribution
-- of this software, copies of which may be obtained from:
--             http://www.cse.ogi.edu/~mpj/thih/
-- 
-----------------------------------------------------------------------------

module Default where
import List( (\\) )
import Type
import Subst
import Pred
import StdPrel

ambig :: [Tyvar] -> [Pred] -> [(Tyvar,[Pred],[Type])]

ambig vs ps
  = [ (v, qs, defs v qs) |
         v <- tv ps \\ vs,
         let qs = [ p | p<-ps, v `elem` tv p ] ]

defs     :: Tyvar -> [Pred] -> [Type]
defs v qs = [ t | all ((TVar v)==) ts,
                  all (`elem` stdClasses) cs,
                  any (`elem` numClasses) cs,
                  t <- defaults,
                  and [ [] ||- IsIn c t | c <- cs ]]
 where cs = [ c | (IsIn c t) <- qs ]
       ts = [ t | (IsIn c t) <- qs ]

useDefaults     :: [Tyvar] -> [Pred] -> [Pred]
useDefaults vs ps
  | any null tss = error "ambiguity"
  | otherwise    = ps \\ ps'
    where ams = ambig vs ps
          tss = [ ts | (v,qs,ts) <- ams ]
          ps' = [ p | (v,qs,ts) <- ams, p<-qs ]

topDefaults     :: [Pred] -> Maybe Subst
topDefaults ps
  | any null tss = Nothing
  | otherwise    = Just (zip vs (map head tss))
    where ams = ambig [] ps
          tss = [ ts | (v,qs,ts) <- ams ]
          vs  = [ v  | (v,qs,ts) <- ams ]

-----------------------------------------------------------------------------
