-----------------------------------------------------------------------------
-- Pred:		Predicates
-- 
-- Part of `Typing Haskell in Haskell', version of September 1, 1999
-- Copyright (c) Mark P Jones and the Oregon Graduate Institute
-- of Science and Technology, 1999
-- 
-- This program is distributed as Free Software under the terms
-- in the file "License" that is included in the distribution
-- of this software, copies of which may be obtained from:
--             http://www.cse.ogi.edu/~mpj/thih/
-- 
-----------------------------------------------------------------------------

module Pred where
import List(union)
import Id
import Type
import Subst
import Unify
import PPrint

data Qual t = [Pred] :=> t
              deriving Eq

instance PPrint t => PPrint (Qual t) where
  pprint (ps  :=> t) = (pprint ps <+> text ":=>") $$ nest 2 (parPprint t)

data Pred   = IsIn Class Type
              deriving Eq

instance PPrint Pred where
  pprint (IsIn c t) = text "isIn1" <+> text ('c' : name c) <+> parPprint t

instance Types t => Types (Qual t) where
  apply s (ps :=> t) = apply s ps :=> apply s t
  tv (ps :=> t)      = tv ps `union` tv t

instance Types Pred where
  apply s (IsIn c t) = IsIn c (apply s t)
  tv (IsIn c t)      = tv t

-----------------------------------------------------------------------------

data Class = Class { name :: Id,
                     super :: [Class],
                     insts :: [Inst] }
type Inst  = Qual Pred

instance Eq Class where
  c == c'  =  name c == name c'

bySuper :: Pred -> [Pred]
bySuper p@(IsIn c t)
 = p : concat (map bySuper supers)
   where supers = [ IsIn c' t | c' <- super c ]

byInst             :: Pred -> Inst -> Maybe [Pred]
byInst p (ps :=> h) = do u <- matchPred h p
                         Just (map (apply u) ps)

matchPred :: Pred -> Pred -> Maybe Subst
matchPred (IsIn c t) (IsIn c' t')
      | c == c'   = match t t'
      | otherwise = Nothing

reducePred :: Pred -> Maybe [Pred]
reducePred p@(IsIn c t) = foldr (|||) Nothing poss
 where poss = map (byInst p) (insts c)
       Nothing ||| y = y
       Just x  ||| y = Just x

-----------------------------------------------------------------------------

(||-)   :: [Pred] -> Pred -> Bool
ps ||- p = any (p `elem`) (map bySuper ps) ||
           case reducePred p of
             Nothing -> False
             Just qs -> all (ps ||-) qs

-----------------------------------------------------------------------------
